# Standard library
import asyncio
import contextlib
import contextvars
import functools
import logging
import threading
from typing import (
    Any,
    Callable,
    cast,
    List,
    Optional,
)

# Local libraries
from tracers.analyzers import (
    analyze_loop_snapshots,
    analyze_stack,
)
from tracers.constants import (
    LOGGER_DEFAULT,
    LOOP_CHECK_INTERVAL,
    T,
)
from tracers.containers import (
    DaemonResult,
    Frame,
    LoopSnapshot,
)
from tracers.contextvars import (
    LEVEL,
    LOGGER,
    STACK,
    TRACING,
)
from tracers.daemon import (
    send_result_to_daemon,
)
from tracers.utils import (
    condition,
    delta,
    divide,
    get_function_id,
    get_monotonic_time,
    increase_counter,
)


def measure_loop_skew(
    should_measure: threading.Event,
    snapshots: List[LoopSnapshot],
) -> None:

    def callback_handler(start_timestamp: float) -> None:
        real_tick_duration: float = \
            delta(start_timestamp, get_monotonic_time())

        snapshots.append(LoopSnapshot(
            block_duration_ratio=divide(
                numerator=real_tick_duration,
                denominator=LOOP_CHECK_INTERVAL,
                on_zero_denominator=1.0,
            ),
            real_tick_duration=real_tick_duration,
            timestamp=start_timestamp,
            wanted_tick_duration=LOOP_CHECK_INTERVAL,
        ))

        schedule_callback()

    def schedule_callback() -> None:
        if should_measure.is_set():
            with contextlib.suppress(RuntimeError):
                asyncio.get_running_loop().call_later(
                    LOOP_CHECK_INTERVAL,
                    callback_handler,
                    get_monotonic_time(),
                )

    schedule_callback()


def record_event(
    event: str,
    function: Callable[..., Any],
) -> None:
    STACK.set(STACK.get() + (Frame(
        event=event,
        function=get_function_id(function),
        level=LEVEL.get(),
        timestamp=get_monotonic_time(),
    ),))


def trace(  # noqa: MC0001
    *,
    enabled: bool = True,
    log_to: Optional[logging.Logger] = LOGGER_DEFAULT,
    overridden_function: Optional[Callable[..., Any]] = None,
) -> Callable[[T], T]:

    def decorator(function: Callable[..., Any]) -> Callable[..., Any]:

        display_function = overridden_function or function

        if asyncio.iscoroutinefunction(function):

            @functools.wraps(function)
            async def isolated_wrapper(*args: Any, **kwargs: Any) -> Any:

                @functools.wraps(function)
                async def wrapper(*args: Any, **kwargs: Any) -> Any:
                    if LEVEL.get() == 0:
                        LOGGER.set(log_to)

                    if enabled and TRACING.get():
                        with condition() as should_measure_loop_skew, \
                                increase_counter(LEVEL):
                            if LEVEL.get() == 1:
                                should_measure_loop_skew.set()
                                snapshots: List[LoopSnapshot] = []
                                measure_loop_skew(
                                    should_measure_loop_skew,
                                    snapshots,
                                )

                            record_event('call', display_function)
                            result = await function(*args, **kwargs)
                            record_event('return', display_function)

                            if LEVEL.get() == 1:
                                stack = STACK.get()
                                analyze_stack(stack)
                                analyze_loop_snapshots(tuple(snapshots))
                                send_result_to_daemon(
                                    result=DaemonResult(
                                        stack=stack,
                                    ),
                                )
                    else:
                        # Disable downstream tracers
                        TRACING.set(False)

                        # No overhead is introduced!
                        result = await function(*args, **kwargs)

                    return result

                return await wrapper(*args, **kwargs)

        elif callable(function):

            @functools.wraps(function)
            def isolated_wrapper(*args: Any, **kwargs: Any) -> Any:

                @functools.wraps(function)
                def wrapper(*args: Any, **kwargs: Any) -> Any:
                    if LEVEL.get() == 0:
                        LOGGER.set(log_to)

                    if enabled and TRACING.get():
                        with increase_counter(LEVEL):
                            record_event('call', display_function)
                            result = function(*args, **kwargs)
                            record_event('return', display_function)

                            if LEVEL.get() == 1:
                                stack = STACK.get()
                                analyze_stack(stack)
                                send_result_to_daemon(
                                    result=DaemonResult(
                                        stack=stack,
                                    ),
                                )
                    else:
                        # Disable downstream tracers
                        TRACING.set(False)

                        # No overhead is introduced!
                        result = function(*args, **kwargs)

                    return result

                if LEVEL.get() == 0:
                    result = contextvars.copy_context().run(
                        wrapper, *args, **kwargs
                    )
                else:
                    result = wrapper(*args, **kwargs)

                return result

        else:

            # We were not able to wrap this object
            raise TypeError(
                f'Expected callable or coroutine function, '
                f'got: {type(function)}'
            )

        return isolated_wrapper

    return cast(Callable[[T], T], decorator)


_DEFAULT_TRACER: Callable[[T], T] = trace()


def call(function: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    return _DEFAULT_TRACER(function)(*args, **kwargs)



# Standard library
import asyncio
import contextlib
from contextvars import (
    ContextVar,
    Token,
)
from decimal import Decimal
import json
import threading
import time
from typing import (
    Any,
    Callable,
    Iterator,
)

# Local libraries
from tracers.containers import (
    DaemonResult,
    Frame,
    LoopSnapshot,
)
from tracers.contextvars import (
    LOGGER,
)


@contextlib.contextmanager
def condition() -> Iterator[threading.Event]:
    instance: threading.Event = threading.Event()
    try:
        yield instance
    finally:
        instance.clear()


def delta(start_timestamp: float, end_timestamp: float) -> float:
    return end_timestamp - start_timestamp


def divide(
    *,
    numerator: float,
    denominator: float,
    on_zero_denominator: float,
) -> float:
    return \
        on_zero_denominator if denominator == 0.0 else numerator / denominator


def get_function_id(function: Callable[..., Any]) -> str:
    # Adding decorators to a function modify its metadata
    #   Fortunately functools' wrapped functions keep a reference to the parent
    while hasattr(function, '__wrapped__'):
        function = getattr(function, '__wrapped__')

    module: str = function.__module__
    name: str = function.__name__
    prefix = 'async ' * asyncio.iscoroutinefunction(function)

    if module not in {'__main__'}:
        return f'{prefix}{module}.{name}'

    return f'{prefix}{name}'


def get_monotonic_time() -> float:
    return time.clock_gettime(time.CLOCK_MONOTONIC)


@contextlib.contextmanager
def increase_counter(contextvar: ContextVar[int]) -> Iterator[None]:
    token: Token[int] = contextvar.set(contextvar.get() + 1)
    try:
        yield
    finally:
        contextvar.reset(token)


def json_dumps(element: object) -> str:

    def cast(obj: Any) -> Any:
        if isinstance(obj, (DaemonResult, Frame, LoopSnapshot)):
            casted_obj: Any = dict(zip(obj._fields, cast(tuple(obj))))
        elif isinstance(obj, (list, set, tuple)):
            casted_obj = list(map(cast, obj))
        elif isinstance(obj, Decimal):
            casted_obj = float(obj)
        else:
            casted_obj = obj

        return casted_obj

    return json.dumps(cast(element))


def log(*parts: str, level: str = 'info') -> None:
    logger = LOGGER.get()
    if logger:
        getattr(logger, level)(' '.join(parts))



# Standard library
import logging
from typing import (
    TypeVar,
)

CHAR_SPACE = chr(0x20)
CHAR_INFO = chr(0x1F6C8) + CHAR_SPACE
CHAR_CHECK_MARK = chr(0X2713)
CHAR_BROKEN_BAR = chr(0xA6)
CHAR_SUPERSCRIPT_ONE = chr(0x00B9)

LOGGER_DAEMON_HANDLER: logging.Handler = logging.StreamHandler()
LOGGER_DAEMON_HANDLER.setLevel(logging.INFO)
LOGGER_DAEMON_HANDLER.setFormatter(
    logging.Formatter('[%(levelname)s] %(message)s')
)
LOGGER_DEFAULT_HANDLER: logging.Handler = logging.StreamHandler()
LOGGER_DEFAULT_HANDLER.setLevel(logging.INFO)
LOGGER_DEFAULT_HANDLER.setFormatter(
    logging.Formatter('%(message)s')
)

LOGGER_DAEMON: logging.Logger = logging.getLogger('Tracers Daemon')
LOGGER_DAEMON.setLevel(logging.INFO)
if not LOGGER_DAEMON.hasHandlers():
    LOGGER_DAEMON.addHandler(LOGGER_DAEMON_HANDLER)

LOGGER_DEFAULT: logging.Logger = logging.getLogger('Tracers')
LOGGER_DEFAULT.setLevel(logging.INFO)
if not LOGGER_DEFAULT.hasHandlers():
    LOGGER_DEFAULT.addHandler(LOGGER_DEFAULT_HANDLER)

LOOP_CHECK_INTERVAL: float = 0.01
LOOP_SKEW_TOLERANCE: float = 1.0

T = TypeVar('T')  # pylint: disable=invalid-name
